<?php
namespace Controller;

use Service\ConnectionService;
use Service\Traits\TResponse;
use Service\DeleteCascadeService;
use Service\HierarchyType;
use Util\Utilities;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Zeedhi\Framework\DTO\Request;
use Zeedhi\Framework\DTO;
use Zeedhi\Framework\DTO\Response;
use Zeedhi\Framework\DTO\Response\Message;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;
use Service\Manager\Mongo\Manager;
use \Service\Hierarchy;
use \Service\UserStructures;
use \Service\DataSource;

class Structure extends Crud {
    
    use TResponse;

    protected $dataSourceName = 'structure';
    /* @var \Service\UserStructures $userStructureService */
    protected $userStructureService;
    /* @var \Service\Structure $structureService */
    protected $structureService;
    /* @var \Service\Hierarchy $hierarchyService */
    protected $hierarchyService;
    /* @var \Service\DataSource $dataSourceService */
    protected $dataSourceService;
    /* @var \Util\Environment $environment */
    protected $environment;

    protected $connectionService;

    public function __construct(Manager $organizationMongoManager, DeleteCascadeService $deleteCascadeService, \Service\Structure $structureService, ConnectionService $connectionService, Hierarchy $hierarchyService, HierarchyType $hierarchyTypeService, DataSource $dataSourceService, \Util\Environment $environment) {
        parent::__construct($organizationMongoManager, $deleteCascadeService);
        $this->connectionService = $connectionService;
        $this->hierarchyService = $hierarchyService;
        $this->structureService = $structureService;
        $this->hierarchyTypeService = $hierarchyTypeService;
        $this->dataSourceService = $dataSourceService;
        $this->environment = $environment;
        $this->mongoManager->createDatabase(\Util\Utilities::getOrganizationIdForMongo());
    }

    public function setUserStructureService(UserStructures $userStructureService) {
        $this->userStructureService = $userStructureService;
    }

  public function find(DTO\Request\Filter $request, DTO\Response $response) {
        $filterData = $request->getFilterCriteria();
        $filterData->setDataSourceName($this->dataSourceName);
        $dataSet = $this->mongoManager->findBy($filterData);
        $decodedStructures = array();
        foreach($dataSet->getRows() as &$structure) {
            $structure['externalIdString'] = Utilities::encodeStructureExternalId($structure['externalId']);
            $decodedStructures[] = $structure;
        }
        $response->addDataSet(new DataSet($this->dataSourceName, $decodedStructures));
    }

    
    public function findNext(DTO\Request $request, DTO\Response $response) {
        try {
            $params = $request->getParameters();    
            $dataSet = $this->structureService->findNext($params);
            $decodedStructures = array();
            foreach($dataSet as &$structure) {
                $structure['externalIdString'] = Utilities::encodeStructureExternalId($structure['externalId']);
                $decodedStructures[] = $structure;
            }
            $dataSet = Utilities::parseDataSourceZeedhiNext($decodedStructures);
            $this->sendResponse($dataSet, 200);
        } catch(\Exception $ex){
            $this->sendResponse(['error' => $ex->getMessage(), 'code' => $ex->getCode()], 500);
        }
    }

    public function count($filter) 
    {
        $filterCriteria = new FilterCriteria($filter->getDataSourceName(), $filter->getConditions());
        $filterCriteria->setPage(1);
        $filterCriteria->setPageSize(9999);
        return $this->mongoManager->count($filterCriteria);
    }

    public function deleteAllStructure(DTO\Request\Filter $request, DTO\Response $response) {
        /** @var \Service\Manager\Mongo\Manager $this->mongoManager */
        $this->mongoManager->dropCollection('structure');
        $this->mongoManager->dropCollection('hierarchyType');
        $this->mongoManager->dropCollection('structureType');
    }

    private function getAllStructureTypeIds(){
        $allStructureTypes = $this->structureService->findAllStructureTypes();
        $structureTypeIds = [];
        foreach($allStructureTypes as $structureType){
            $structureTypeIds[] = $structureType["_id"];
        }
        return $structureTypeIds;
    }

    public function findStructureTree(DTO\Request\Filter $request, DTO\Response $response) {
        $filterData = $request->getFilterCriteria();
        $filterData->setDataSourceName($this->dataSourceName);
        $structureTypeIds = $this->getAllStructureTypeIds();
        $filterData->addCondition('type', FilterCriteria::IN, $structureTypeIds);
        $dataSet = $this->mongoManager->findBy($filterData);
        $decodedStructures = array();
        foreach($dataSet->getRows() as &$structure) {
            $structure['externalIdString'] = Utilities::encodeStructureExternalId($structure['externalId']);
            $obj = array("columnName"=> "structureChild",
                "value"=>$structure["_id"],
                "operator"=> "=" );
            $filterHierarchy = new FilterCriteria("hierarchy",[$obj]);
            $dataSetHierarchy = $this->mongoManager->findBy($filterHierarchy);
            if(isset($dataSetHierarchy->getRows()[0])){
                $structure["parent"] = $dataSetHierarchy->getRows()[0]["structureParent"];
            }else{
                $structure["parent"] = null;
            }
            $decodedStructures[] = $structure;
        }
        $response->addDataSet(new DataSet($this->dataSourceName, $decodedStructures));
    }
    
    public function delete(Request $request, Response $response) {
        parent::delete($request, $response);
        foreach($request->getDataSet()->getRows() as $row){
            $newRow = $this->structureService->findById($row["_id"]);
            $this->structureService->saveLogStructure([$row], [$newRow], $row["_id"], $row["name"], $row["type"]);
        }
    }

    public function save(DTO\Request $request, Response $response) {
        try {
            $request = $this->buildDataSetIfRow($request);
            $requestRows = $request->getDataSet()->getRows();
            $oldRows = [];
            if(!empty($this->structureService->findStructureTypeById($requestRows[0]["type"]))){
                foreach($requestRows as $i => $row) {
                    if(isset($rows[$i]['externalId'])) {
                        $rows[$i]['externalId'] = Utilities::decodeStructureExternalId($rows[$i]['externalId']);
                    }
                    if(isset($rows[$i]['CDOPERADOR'])) {
                        unset($rows[$i]['CDOPERADOR']);
                    }
                    if(isset($rows[$i]['NMOPERADOR'])) {
                        unset($rows[$i]['NMOPERADOR']);
                    }
                    if(!empty($row["_id"])){
                        $oldRows[$row["_id"]] = $this->structureService->findById($row["_id"]);
                    }
                }
                parent::save($request, $response);
                $idsDataSets = $response->getDataSets();
                $idsDataSet = $idsDataSets[0];
                $idsRows = $idsDataSet->getRows();
                foreach($request->getDataSet()->getRows() as $index => &$row) {
                    if(isset($idsRows[$index]) && $idsRows[$index]['_id']) {
                        $row['_id'] = $idsRows[$index]['_id'];
                    }
                    $dataBefore = !empty($oldRows[$row["_id"]]) ? $oldRows[$row["_id"]] : null;
                    $dataAfter = $this->structureService->findById($row["_id"]);
                    $this->structureService->saveLogStructure([$dataBefore], [$dataAfter], $dataAfter["_id"], $dataAfter["name"], $dataAfter["type"]);
                }
                $this->userStructureService->persistRelations($request->getDataSet());
            } else {
                throw new \Exception("NONEXISTENT_STRUCTURE");
            }
        } catch (\Exception $e) {
            $response->setError(new Response\Error($e->getMessage(), $e->getCode()));
        }
    }

    private function setParentTo(&$structureHierarchy, $parent, $newParent) {
        foreach ($structureHierarchy as $key => $structure) {
            if($structure['parent'] === $parent) {
                $structureHierarchy[$key]['parent'] = $newParent;
            }
        }
    }

    private function findParent($structureHierarchy, $parent) {
        foreach ($structureHierarchy as $key => $structure) {
            if($structure['_id'] === $parent) {
                return $structure;
            }
        }
    }

    public function getUserStructuresByIndicator(Request\Filter $request, Response $response) {
        try {
            $filterCriteria = $request->getFilterCriteria()->getConditions();
            $indicator = $filterCriteria[0]['value'];
            $userStructureHierarchy = $this->hierarchyService->getIndicatorHierarchyByLoggedUser($indicator);
            $response->addDataSet(new DataSet('getUserStructuresByIndicator', $userStructureHierarchy));
        } catch (\Exception $e) {
            $response->setError(new Response\Error($e->getMessage(), $e->getCode()));
        }
    }

    public function getStructureByUser(Request\Filter $request, Response $response) {
        try {
            $hierarchy = $this->hierarchyService->getUserHierarchysTrees();
            $response->addDataSet(new DataSet('getStructureByUser', $hierarchy));
        } catch (\Exception $e) {
            $response->setError(new Response\Error($e->getMessage(), $e->getCode()));
        }
    }

    public function loadManyStructuresFromQuery(DTO\Request $request, Response $response){
        $params = $request->getParameters();
        $return = array();

        foreach($params as $param){
            
            $dataSet = new \Zeedhi\Framework\DataSource\FilterCriteria($this->dataSourceName,array(array("columnName"=>"type","operator"=>"=","value"=>$param["_id"])),null,null);
            //$dataSet = array(array("columnName"=>"type","operator"=>"=","value"=>$param["_id"]), null, null);
            $data = $this->mongoManager->findBy($dataSet)->getRows();
            $query = $param['query'];
            $organizationId = Utilities::getOrganizationIdForMongo();
            $clientRoom = $this->connectionService->getClientRoom($organizationId, $param);
            $connectionId = $param["connection"] ?: null;
            $rows = $this->dataSourceService->execQueryFromWebSocket($query, new FilterCriteria($this->dataSourceName),
                $clientRoom, 7, $connectionId,"","",false);
            $fieldKey = $param['fieldKey'];
            $descriptionField = $param['descriptionField']; 
            $bdStructures = array();
            foreach ($rows as $row) {
                $bdStructures[] = $this->structureService->generateStructureFromQueryData($row, $fieldKey, $descriptionField, $param['_id']);
            }
            $newStructures = $this->structureService->getNewStructures($bdStructures, $data);
            foreach($newStructures as $newStructure){
                $newStructure['typeName'] = $param['name'];
                $newStructure['externalIdString'] = Utilities::encodeStructureExternalId($newStructure['externalId']);
                array_push($return,$newStructure);
            }
        }
        $this->sendResponse($return, 200); 
    }

    public function loadStructuresFromQuery(Request\Filter $request, Response $response) {
        $filterCriteria = $request->getFilterCriteria();
        $structureType = $filterCriteria->getConditions()[0]['value'];
        $oldStructures = $filterCriteria->getConditions()[1]['value'];
        
        try {
            $newStructures = $this->structureService->loadStructuresFromQuery($structureType,$oldStructures);
            $dataSet = new DataSet($this->dataSourceName, $newStructures);
            $response->addDataSet($dataSet);
        } catch (Exception $e) {
            if($e->getCode() === Exception::CODE_FAIL_CONNECTION || $e->getCode() === Exception::CODE_QUERY_EXECUTION_FAIL) {
                $response->addMessage(new Message($e->getMessage(), Message::TYPE_ERROR));
            } else {
                $response->setError(new Response\Error($e->getMessage(), $e->getCode()));
            }
        }
    }

    public function getStructDrillInfoByName(Request\Row $row, Response $response) {
        try {
            $row = $row->getRow();
            $childStructure = null;
            $drillDownColumn = null;
            $structure = null;
            if (!empty($row['structureName'])) {
                $filterCriteria = new FilterCriteria($this->dataSourceName);
                $filterCriteria->addCondition('name', $row['structureName']);
                $structure = $this->structureService->findBy($filterCriteria);
                if (!empty($structure)) {
                    $structure = $structure[0];
                    $childStructure = $this->hierarchyTypeService->getChildStructureTypeByStructureType($structure['type']);
                    if (!empty($row['indicatorId'])) {
                        $drillDownColumn = $this->getDrillDownColumnFromStructureAndIndicatorId($structure, $row['indicatorId']);
                    }
                }
            }

            $result = array(
                'childStructure' => $childStructure,
                'structure' => $structure,
                'drillDownColumn' => $drillDownColumn
            );

            $response->addDataSet(new DataSet('structureInfo', $result));
        } catch (\Exception $e) {
            $response->setError(new Response\Error($e->getMessage(), $e->getCode()));
        }
    }

    /**
     * We will find out what is the name of the column to be used in the drill down from the structureType descriptionField
     * If it's not available, the association between the structureType's fieldKey and the dataSource relations will be used
     * @param array $structure
     * @param array $indicatorId
     * @return $drillDownColumn|null
     */
    protected function getDrillDownColumnFromStructureAndIndicatorId(array $structure, $indicatorId) {
        $drillDownColumn = null;
        $structureType = $this->structureService->findStructureTypeById($structure['type']);
        if($structureType) {
            if(!empty($structureType['descriptionField'])) {
                $drillDownColumn = $structureType['descriptionField'];
            } else {
                $dataSource = $this->dataSourceService->findByIndicator($indicatorId);
                if($dataSource && !empty($dataSource['relatedStructuresType'])) {
                    foreach ($dataSource['relatedStructuresType'] as $relatedStructureType) {
                        if($relatedStructureType['structureType'] === $structure['type']) {
                            foreach ($relatedStructureType['relations'] as $relation) {
                                if(in_array($relation['structureColumn'], $structureType['fieldKey'])) {
                                    $drillDownColumn = $relation['dataSourceColumn'];
                                }
                            }
                        }
                    }
                }
            }
        }

        return $drillDownColumn;
    }
}