<?php
namespace Service;

use Service\Manager\Mongo\Manager;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DataSource\FilterCriteria;
use Util\Environment;
use Util\Utilities;

class DeleteCascadeService {

    const CONFIGURATION_PROPERTY = 'property';
    const CONFIGURATION_CASCADE = 'cascade';
    const CONFIGURATION_SUBLEVEL = 'subLevel';
    const CONFIGURATION_PARENTS_IDS =  'parentsIds';

    const MAP_COLLECTIONS = 'collections';
    const MAP_COLLECTION_NAME = 'name';
    const MAP_FKS = 'foreignKeys';
    const MAP_FK_COLLECTION_NAME = 'collectionName';
    const MAP_COLLECTION_KEYS = 'keys';
    const MAP_KEY_NAME = 'name';

    protected $organizationManager;
    protected $environment;
    protected $dataBaseMap;

    public function __construct(Manager $organizationManager, Environment $environment) {
        $this->organizationManager = $organizationManager;
        $this->environment = $environment;
        $this->dataBaseMap = $this->getDataBaseMap();
    }

    public function deleteCascade(DataSet $dataSet, $previousConfiguration = null) {
        if($this->hasSubLevel($previousConfiguration)) {
            $this->subLevelDelete(
                $dataSet,
                $this->getConfigurationProperty($previousConfiguration),
                $this->isCascade($previousConfiguration),
                $this->getConfigurationParentsId($previousConfiguration)
            );
        } else {
            if(is_null($previousConfiguration) || $this->isCascade($previousConfiguration)) {
                $this->cascade(
                    $this->findCollectionConfiguration($dataSet->getDataSourceName()),
                    $this->getDataSetIds($dataSet)
                );
                $this->realDelete($dataSet);
            } else {
                $this->setNull($dataSet, $this->getConfigurationProperty($previousConfiguration));
            }
        }
    }

    protected function cascade($configuration, $documentsIds) {
        if(!is_null($configuration)) {
            foreach($configuration[self::MAP_FKS] as $fkConfiguration) {
                $collectionName = $fkConfiguration[self::MAP_FK_COLLECTION_NAME];
                foreach($fkConfiguration[self::MAP_COLLECTION_KEYS] as $keyConfiguration) {
                    $postConfiguration = $this->factoryPostConfiguration($keyConfiguration, $documentsIds);

                    $filterCriteria = $this->createFilterCriteria($collectionName, $keyConfiguration[self::MAP_KEY_NAME], $documentsIds, FilterCriteria::IN);
                    $dependents = $this->findBy($filterCriteria);

                    if (sizeof($dependents->getRows()) > 0) {
                        $this->deleteCascade($dependents, $postConfiguration);
                    }
                }
            }
        }
    }

    protected function realDelete (DataSet $dataSet) {
        $this->organizationManager->delete($dataSet);
    }

    protected function setNull (DataSet $dataSet, $property) {
        $rows = $dataSet->getRows();
        foreach($rows as &$row) {
            $row[$property] = null;
        }
        $this->organizationManager->persist($dataSet);
    }

    protected function subLevelDelete (DataSet $dataSet, $property, $cascade, $keys) {
        $deepProperties = explode(".", $property);
        $documents = $dataSet->getRows();
        $this->searchSubLevelAndDelete($cascade, $keys, $deepProperties, $documents);

        $this->organizationManager->persist($dataSet);
    }

    protected function searchSubLevelAndDelete ($cascade, $keys, $deepProperties, &$document, $deepPropertiesIndex = 0, &$parentDocument = null, $documentIndex = null) {
        if (is_array($document)) {
            $property = $deepProperties[$deepPropertiesIndex];

            if (array_key_exists($property, $document)) {
                if ($this->isFinalLevel($deepPropertiesIndex, $deepProperties)) {
                    $this->concretizeSubLevelDelete($cascade, $keys, $property, $document, $parentDocument, $documentIndex);
                } else {
                    $this->searchSubLevelAndDelete($cascade, $keys, $deepProperties, $document[$property], ++$deepPropertiesIndex, $document, $property);
                }
            } else if ($this->isIndexedArray($document)) {
                foreach($document as $index => &$item) {
                    $this->searchSubLevelAndDelete($cascade, $keys, $deepProperties, $item, $deepPropertiesIndex, $document, $index);
                }
            }
        }
    }

    protected function concretizeSubLevelDelete ($cascade, $keys, $property, &$document, &$parentDocument, $documentIndex) {
        if (in_array($document[$property], $keys)) {
            if ($cascade) {
                unset($parentDocument[$documentIndex]);
            } else {
                $document[$property] = null;
            }
        }
    }

    protected function isFinalLevel ($index, $deepProperties) {
        return $index === (sizeof($deepProperties) - 1);
    }

    protected function isIndexedArray ($array) {
        return !(Utilities::array_some(array_keys($array), 'is_string'));
    }

    protected function factoryPostConfiguration($keyConfiguration, $parentsIds) {
        return array(
            self::CONFIGURATION_CASCADE =>  $this->isCascade($keyConfiguration),
            self::CONFIGURATION_PROPERTY => $keyConfiguration[self::MAP_KEY_NAME],
            self::CONFIGURATION_SUBLEVEL => $this->isSubLevel($keyConfiguration[self::MAP_KEY_NAME]),
            self::CONFIGURATION_PARENTS_IDS => $parentsIds
        );
    }

    protected function findCollectionConfiguration ($collectionName) {
        foreach($this->dataBaseMap[self::MAP_COLLECTIONS] as $collection) {
            if ($this->hasProperty($collection, self::MAP_COLLECTION_NAME) && $collection[self::MAP_COLLECTION_NAME] === $collectionName) {
                return $collection;
            }
        }
        return null;
    }

    protected function createFilterCriteria ($collectionName, $property, $ids, $operator = "=") {
        $filterCriteria = new FilterCriteria($collectionName);
        $filterCriteria->addCondition($property, $operator, $ids);
        $filterCriteria->setPageSize(0);
        return $filterCriteria;
    }

    protected function findBy(FilterCriteria $filterCriteria) {
        return $this->organizationManager->findBy($filterCriteria);
    }

    protected function getDataSetIds(DataSet $dataSet) {
        $rows = $dataSet->getRows();

        $ids = array();
        foreach($rows as $row) {
            $ids[] = $row['_id'];
        }
        return $ids;
    }

    protected function getPropertyFromArray ($configuration, $property) {
        return $this->hasProperty($configuration, $property) ? $configuration[$property] : false;
    }

    protected function getConfigurationProperty($configuration) {
        return $this->getPropertyFromArray($configuration, self::CONFIGURATION_PROPERTY);
    }

    protected function getConfigurationParentsId($configuration) {
        return $this->getPropertyFromArray($configuration, self::CONFIGURATION_PARENTS_IDS);
    }

    protected function hasSubLevel($configuration) {
        return $this->getPropertyFromArray($configuration, self::CONFIGURATION_SUBLEVEL);
    }

    protected function isCascade($configuration) {
        return $this->getPropertyFromArray($configuration, self::CONFIGURATION_CASCADE);
    }

    protected function isSubLevel($property) {
        return strpos($property, ".") !== false;
    }

    protected function hasProperty($array, $property) {
        return isset($array[$property]);
    }

    protected function getDataBaseMap () {
        return json_decode(file_get_contents($this->environment->getDataBaseMapFilePath()), true);
    }

}