<?php

namespace Service;

use Zeedhi\Framework\DataSource\FilterCriteria;
use \Service\UserStructure;
use \Service\Indicator;
use \Service\Manager\Mongo\Manager;
use Zeedhi\Framework\DTO\Row;
use Util\Utilities;

class Hierarchy {

    protected $dataSourceName = 'hierarchy';
    /** @var \Service\Manager\Mongo\Manager $organizationManager */
    protected $organizationManager;
    /** @var \Service\HierarchyType $hierarchyType */
    protected $hierarchyTypeService;
    /** @var \Service\Structure $structureService */
    protected $structureService;
    /* @var \Service\UserStructures $userStructureService */
    protected $userStructureService;
    /* @var \Service\Indicator $indicatorService */
    protected $indicatorService;

    public function __construct(Manager $organizationManager, HierarchyType $hierarchyTypeService, Structure $structureService, Indicator $indicatorService)
    {
        $this->organizationManager = $organizationManager;
        $this->hierarchyTypeService = $hierarchyTypeService;
        $this->structureService = $structureService;
        $this->indicatorService = $indicatorService;
    }

    public function findNext ($params) {
        $response = $this->organizationManager->findNext($this->dataSourceName, $params);
        return $response;
    }

    public function getStrucutresTypes ($hierarchy) {
        $usedIds = $this->getStructureIdsUsed($hierarchy);
        $numberOfExpectedLines = count($usedIds);
        $filterCriteria = new FilterCriteria('structureType', array(), null, $numberOfExpectedLines);
        $filterCriteria->addCondition('_id', FilterCriteria::IN, $usedIds);
        return $this->organizationManager->findBy($filterCriteria);
    }

    public function generateHierarchy ($parentStructure, &$childsStructures, $hierarchyType, $fieldKeysRelations) {
        $hierarchyRows = array();
        foreach ($childsStructures as $key => $childStructure) {
            if ($this->isChild($parentStructure, $childStructure, $fieldKeysRelations)) {
                $hierarchy = $this->uniteParentChild($parentStructure, $childStructure);
                $hierarchyRows[] = $this->convertHierarchyInRow($hierarchy, $hierarchyType);
                unset($childsStructures[$key]);
            }
        }
        return $hierarchyRows;
    }

    public function indexByMongoId($rows) {
        $newRows = array();
        foreach($rows as $row) {
            $newRows[$row['_id']] = $row;
        }
        return $newRows;
    }

    protected function uniteParentChild ($parentStructure, $childStructure) {
        return array(
            "structureParent" => $parentStructure,
            "structureChild" => $childStructure
            );
    }

    protected  function convertHierarchyInRow($hierarchy, $type) {
        return new Row(array(
            "name" => $this->generateHierarchyName($hierarchy),
            "structureParent" => $hierarchy['structureParent']['_id'],
            "structureChild" => $hierarchy['structureChild']['_id'],
            "type" => $type,
            "__is_new" => true,
            "changed" => true
            ));
    }

    protected function generateHierarchyName ($hierarchy) {
        return "{$hierarchy['structureParent']['name']} - {$hierarchy['structureChild']['name']}";
    }

    protected function isChild ($parentStructure, $childStructure, $fieldKeysRelations) {
        $parentExternalId = Utilities::decodeStructureExternalId($parentStructure['externalId']);
        $childExternalId = Utilities::decodeStructureExternalId($childStructure['externalId']);
        foreach($fieldKeysRelations as $fieldKeysRelation) {
            if (!isset($childExternalId[$fieldKeysRelation['childKey']]) || !isset($parentExternalId[$fieldKeysRelation['parentKey']])
                || $childExternalId[$fieldKeysRelation['childKey']] != $parentExternalId[$fieldKeysRelation['parentKey']]) {
                return false;
            }
        }
        return true;
    }

    public function relateHierarchyWithStructureType($hierarchies, $indexedByIdStructure) {
        foreach($hierarchies as &$hierarchyType) {
            $hierarchyType['structureParentName'] = $indexedByIdStructure[$hierarchyType['structureParent']]['name'];
            $hierarchyType['structureChildName'] = $indexedByIdStructure[$hierarchyType['structureChild']]['name'];
        }
        return $hierarchies;
    }

    public function relateHierarchyWithHierarchyParent($hierarchys, $indexedByIdHierarchy) {
        foreach($hierarchys as &$hierarchy) {
            if (isset($hierarchy['parent'])) {
                $hierarchy['parentName'] = $indexedByIdHierarchy[$hierarchy['parent']]['name'];
            }
        }
        return $hierarchys;
    }

    public function getStructureIdsUsed ($hierarchies) {
            //For not repeat Ids used more than one time
        $structuresIdsIndexedById = $this->getDistinctStructureIds($hierarchies);
            //Build distinct ids in a index array
        $distictStruturesIds = array();
        foreach($structuresIdsIndexedById as $structureId) {
            $distictStruturesIds[] = $structureId;
        }
        return $distictStruturesIds;
    }

    protected function getDistinctStructureIds($hierarchies) {
        $structuresIdsIndexedById = array();
        foreach($hierarchies as $hierarchyType) {
            $structureParentId = $hierarchyType['structureParent'];
            $structuresIdsIndexedById[$structureParentId] = $structureParentId;
            $structureChildId = $hierarchyType['structureChild'];
            $structuresIdsIndexedById[$structureChildId] = $structureChildId;
        }
        return $structuresIdsIndexedById;
    }



    public function setUserStructureService(UserStructures $userStructureService) {
        $this->userStructureService = $userStructureService;
    }

    public function findBy(FilterCriteria $filterCriteria) {
        return $this->organizationManager->findBy($filterCriteria)->getRows();
    }

    public function findById($id) {
        $filterCriteria = new FilterCriteria($this->dataSourceName, array(array("columnName"=>"_id", "value" => $id, "operator" => "=")), null, 10000);
        $structure = $this->organizationManager->findBy($filterCriteria)->getRows();
        return count($structure) > 0 ? $structure[0] : null;
    }

    public function findAll() {
        return $this->organizationManager->findBy(new FilterCriteria($this->dataSourceName, array(), null, 10000))->getRows();
    }

    private function buildAllHierarchyTrees($hierarchyTypes, $hierarchies, $structures, $parent = null, &$trees = array()) {
        $trees = [];
        if(!empty($hierarchyTypes) || !empty($hierarchies) || !empty($structures)) {
            $indexedHierarchyTypes = $this->indexBy('typeChild', $hierarchyTypes);
            $hierarchiesBy = $this->aggregateBy($hierarchies, $indexedHierarchyTypes, 'type', '_id', "hierarchies");

            foreach ($structures as &$structure) {
                $type = $structure['type'];
                if (isset($hierarchiesBy[$type])) {
                    $this->matchChild($structure, $hierarchiesBy[$type]);
                } else {
                    $structure['parent'] = null;
                }
            }
            $trees = $structures;
        }
        return $trees;
    }

    private function matchChild (&$structure, &$hierarchyType) {
        if (!isset($hierarchyType["hierarchies"])) return;
        foreach ($hierarchyType["hierarchies"] as $key => &$hierarchy) {
            if($hierarchy['structureChild'] == $structure['_id']) {
                $structure['parent'] = $hierarchy['structureParent'];
                unset($hierarchyType["hierarchies"][$key]);
            }
        }
    }

    private function indexBy($index, $rows){
        $indexedRows = array();
        foreach ($rows as $row) {
            $indexedRows[$row[$index]] = $row;
        }
        return $indexedRows;
    }

    private function findInArray($value, $property, &$array){
        $found = array();
        foreach ($array as $index => $item) {
            if (isset($item[$property]) && $item[$property] == $value) {
                $found[] = $item;
                unset($array[$index]);
            }
        }
        return $found;
    }

    private function aggregateBy ($array, $aggregators, $commonProperty, $aggregateProperty, $category = "child") {
        $aggregateArray = $aggregators;
        foreach ($aggregators as $index => $aggregator) {
            foreach ($array as $key => $item) {
                if (isset($item[$commonProperty]) && $item[$commonProperty] == $aggregator[$aggregateProperty]) {
                    $aggregateArray[$index][$category][] = $item;
                    unset($array[$key]);
                }
            }
        }
        return $aggregateArray;
    }

    public function getAllHierarchyTrees($structuresTypes = null) {
        $hierarchyTypes = $this->hierarchyTypeService->findAll();

        if(!empty($structuresTypes)){
            $structures = $this->structureService->findByStructureTypes($structuresTypes);
        } else {
            $structures = $this->structureService->findAll();
        }
        $hierarchies = $this->findAll();

        return $this->buildAllHierarchyTrees($hierarchyTypes, $hierarchies, $structures);
    }

    public function getAllHierarchyTreesPlanned() {
        return $this->getAllHierarchyTrees();
    }

    private function buildUserTrees($hierarchyTrees, $userStructures, &$userHierarchy = array(), $parentId = null) {
        foreach ($hierarchyTrees as $index => $node) {
            if (array_key_exists('parent', $node) && $node['parent'] == $parentId) {
                unset($hierarchyTrees[$index]);
                if (in_array($node['_id'], $userStructures)) {
                    $node['parent'] = null;
                    $userHierarchy[] = $node;
                    $children = $this->getChildrenTree($hierarchyTrees, $node['_id']);
                    $userHierarchy = array_merge($userHierarchy, $children);
                }
                else{
                    $this->buildUserTrees($hierarchyTrees, $userStructures, $userHierarchy, $node['_id']);
                }
            }
        }
        return $userHierarchy;
    }

    private function getChildrenTree (&$hierarchies, $parentId = null, &$childrensTree = array()) {
        $newChildren = $this->findInArray($parentId, 'parent', $hierarchies);
        if (!empty($newChildren)) {
            $childrensTree = array_merge($childrensTree, $newChildren);
            foreach ($newChildren as $parent) {
                $this->getChildrenTree($hierarchies, $parent['_id'], $childrensTree);
            }
        }
        return $childrensTree;
    }


    public function getUserHierarchysTrees($userId = null, $structuresType = null) {
        if(is_null($userId)) {
            $userStructures = $this->userStructureService->getLoggedUserStructures($structuresType);
            $profileStructures = $this->userStructureService->getLoggedUserStructuresAccessProfile();
        } else {
            $userStructures = $this->userStructureService->getUserStructuresByUserId($userId);
        }
        $userStructuresId = array();
        foreach ($userStructures as $structure) {
            $userStructuresId[] = $structure['_id'];
        }
        if(isset($profileStructures)){
            foreach($profileStructures as $profileStructure){
                foreach($profileStructure as $profile){
                    $userStructuresId[] = $profile['_id'];
                }
            }
        }
        $hierarchyTrees = $this->getAllHierarchyTrees($structuresType);
        $hierarchyTree = $this->buildUserTrees($hierarchyTrees, $userStructuresId);
        return $hierarchyTree;
    }

    public function getUserHierarchyTreesPlanned($userId = null) {
        return $this->getUserHierarchysTrees($userId);
    }

    public function getUserParentHierarchysId($userId = null, $structureTypes = null) {
        $userHierarchyTrees = $this->getUserHierarchysTrees($userId,$structureTypes);
        $structuresId = array();
        foreach ($userHierarchyTrees as $structure) {
            $structuresId[] = $structure['_id'];
        }
        return $structuresId;
    }

    protected function setParentTo(&$structureHierarchy, $parent, $newParent) {
        foreach ($structureHierarchy as $key => $structure){
            if(isset($structure['parent']) && $structure['parent'] === $parent) {
                $structureHierarchy[$key]['parent'] = $newParent;
            }
        }
    }

    protected function findParent($structureHierarchy, $parent) {
        foreach ($structureHierarchy as $key => $structure) {
            if($structure['_id'] === $parent) {
                return $structure;
            }
        }
    }
    public function getDataSourceHierarchyByLoggedUser ($dataSourceId, $userStructureHierarchy = null) {

        $indicatorRelatedStructuresType = $this->indicatorService->getRelatedStructuresTypeIdsByDataSource($dataSourceId, true);
        $userStructureHierarchy = $this->getUserHierarchysTrees(null,$indicatorRelatedStructuresType);

        $userStructureHierarchy = $this->validStructure($userStructureHierarchy, $indicatorRelatedStructuresType);

        return array_values($userStructureHierarchy);
    }

    public function getIndicatorHierarchyByLoggedUser ($indicatorId, $userStructureHierarchy = null) {
        $indicatorRelatedStructuresType = $this->indicatorService->getRelatedStructuresTypeIds($indicatorId, true);
        $userStructureHierarchy = $this->getUserHierarchysTrees(null,$indicatorRelatedStructuresType);

        $userStructureHierarchy = $this->validStructure($userStructureHierarchy, $indicatorRelatedStructuresType);

        return array_values($userStructureHierarchy);
    }

    public function getStructureHierarchy($structure) {
        $filterCriteria = $this->buildCriteria('structureParent', $structure);
        $hierarchies = $this->organizationManager->findBy($filterCriteria)->getRows();

        $childrenIds = function($item) {
            return $item['structureChild'];
        };

        $children = array_map($childrenIds, $hierarchies);
        $hierarchy = array('children' => $children);

        $count = 1;
        $structureToSearch = $structure;
        do {
            $parent = $this->getStructureParent($structureToSearch);
            $hierarchy['parent' . $count] = $parent;
            if($parent) {
                $structureToSearch = $parent;
                $count ++;
            } else {
                $count = null;
            }

        } while ($count);

        return $hierarchy;
    }

    protected function buildCriteria($columnName, $value, $operator = '=') {
        return new FilterCriteria($this->dataSourceName, array( array(
            "columnName" => $columnName,
            "value" => $value,
            "operator" => $operator
        )), null, 10000);
    }

    protected function getStructureParent($structureTypeId) {
        $parent = null;
        $filterCriteria = $this->buildCriteria('structureChild', $structureTypeId);
        $hierarchies = $this->organizationManager->findBy($filterCriteria)->getRows();
        $parent = count($hierarchies) > 0 ? $hierarchies[0]['structureParent'] : null;

        return $parent;
    }

    public function validStructure($userStructureHierarchy, array $indicatorRelatedStructuresType)
    {
        foreach ($userStructureHierarchy as $key => $structure) {
            if (!in_array($structure['type'], $indicatorRelatedStructuresType)) {
                $this->setParentTo($userStructureHierarchy, $structure['_id'], $structure['parent']);
                unset($userStructureHierarchy[$key]);
            } else {
                $parent = null;
                if (isset($structure['parent'])) {
                    $parent = $this->findParent($userStructureHierarchy, $structure['parent']);
                }
                if (empty($parent)) {
                    $userStructureHierarchy[$key]['parent'] = null;
                }
            }
        }
        return $userStructureHierarchy;
    }
}
